"""
Test Webdriver steps.
"""

from aloe.testing import FeatureTest

from aloe_webdriver.tests.base import feature, skip_if_browser

# pylint:disable=line-too-long


class TestSteps(FeatureTest):
    """Test steps."""

    @feature()
    def test_I_should_see(self):
        """
        When I visit test page "basic_page"
        Then I should see "Hello there!"
        And I should see 'Welcome "User"'
        And I should see a link to "Google" with the url "http://google.com/"
        And I should see a link with the url "http://google.com/"
        And I should not see "Bogeyman"
        """

    @feature()
    def test_I_see_a_link(self):
        """
        When I visit test page "basic_page"
        Then  I should see a link to "Google" with the url "http://google.com/"
        And I see "Hello there!"
        """

    @feature()
    def test_see_a_link_containing(self):
        """
        When I visit test page "basic_page"
        Then The browser's URL should contain "http://"
        And I should see a link that contains the text "Goo" and the url "http://google.com/"
        """

    @feature()
    def test_basic_page_linking(self):
        """
        Given I visit test page "link_page"
        And I see "Page o link"
        When I click "Next Page"
        Then I should be at "http://SERVER_HOST:7755/link_dest.html"
        And The browser's URL should be "http://SERVER_HOST:7755/link_dest.html"
        And The browser's URL should not contain "irrelevant things"
        And I should see "Link destination page"
        """

    @feature()
    def test_ajax_action(self):
        """
        Given I visit test page "link_page"
        When I click "Load content with AJAX"
        Then I should see "Loaded with AJAX"
        """

    @feature()
    def test_I_see_a_form(self):
        """
        When I visit test page "basic_page"
        Then I should see a form that goes to "basic_page.html"
        And the element with id of "somediv" contains "Hello"
        And the element with id of "somediv" does not contain "bye"
        """

    @feature()
    def test_I_fill_in_a_form(self):
        """
        Given I visit test page "basic_page"
        And I fill in "bio" with "everything awesome"
        And I fill in "Password: " with "neat"
        When I press "Submit!"
        Then The browser's URL should contain "bio=everything"
        """

    @feature()
    def test_checkboxes_checked(self):
        """
        Given I visit test page "basic_page"
        When I check "I have a bike"
        Then The "I have a bike" checkbox should be checked
        And The "I have a car" checkbox should not be checked
        """

    @feature()
    def test_checkboxes_unchecked(self):
        """
        Given I visit test page "basic_page"
        And I check "I have a bike"
        And The "I have a bike" checkbox should be checked
        When I uncheck "I have a bike"
        Then The "I have a bike" checkbox should not be checked
        """

    @feature()
    def test_combo_boxes(self):
        """
        Given I visit test page "basic_page"
        Then I should see option "Mercedes" in selector "car_choice"
        And I should see option "Volvo" in selector "car_choice"
        And I should not see option "Skoda" in selector "car_choice"
        When I select "Mercedes" from "car_choice"
        Then The "Mercedes" option from "car_choice" should be selected
        """

    @feature(fails=True)
    def test_combo_boxes_fail(self):
        """
        Given I visit test page "basic_page"
        Then I should not see option "Mercedes" in selector "car_choice"
        """

    @feature()
    def test_multi_combo_boxes(self):
        '''
        Given I visit test page "basic_page"
        When I select the following from "Favorite Colors:":
            """
            Blue
            Green
            """
        Then The following options from "Favorite Colors:" should be selected:
            """
            Blue
            Green
            """
        '''

    @feature()
    def test_radio_buttons(self):
        """
        When I visit test page "basic_page"
        And I choose "Male"
        Then The "Male" option should be chosen
        And The "Female" option should not be chosen
        """

    @feature(fails=True)
    def test_hidden_text(self):
        """
        When I visit test page "basic_page"
        Then I should see an element with id of "bio_field"
        And I should see an element with id of "somediv"
        And I should not see an element with id of "hidden_text"
        And I should see "Weeeee"
        """

    @feature(fails=True)
    def test_hidden_text_2(self):
        """
        When I visit test page "basic_page"
        Then I should see "Hello there"
        And I should see an element with id of "oops_field"
        And I should not see an element with id of "hidden_text"
        """

    @skip_if_browser('phantomjs', "PhantomJS doesn't support alerts")
    @feature()
    def test_alert_accept(self):
        """
        When I visit test page "alert_page"
        Then I should see an alert with text "This is an alerting alert"
        When I accept the alert
        Then I should not see an alert
        And I should see "true"
        """

    @skip_if_browser('phantomjs', "PhantomJS doesn't support alerts")
    @feature()
    def test_alert_dismiss(self):
        """
        When I visit test page "alert_page"
        Then I should see an alert with text "This is an alerting alert"
        When I dismiss the alert
        Then I should not see an alert
        And I should see "false"
        """

    @feature()
    def test_tooltips(self):
        """
        When I visit test page "tooltips"
        Then I should see an element with tooltip "A tooltip"
        And I should not see an element with tooltip "Does not exist"
        And I should not see an element with tooltip "Hidden"
        When I click the element with tooltip "A tooltip"
        Then the browser's URL should contain "#anchor"
        """

    @feature()
    def test_labels(self):
        """
        When I visit test page "basic_page"
        And I click on label "Favorite Colors:"
        Then element with id "fav_colors" should be focused
        And element with id "bio_field" should not be focused
        """

    @feature(fails=True)
    def test_labels_fail(self):
        """
        When I visit test page "basic_page"
        And I click on label "Favorite Colors:"
        Then element with id "fav_colors" should not be focused
        """

    @feature()
    def test_input_values(self):
        """
        When I visit test page "basic_page"
        And I fill in "username" with "Danni"
        Then input "username" has value "Danni"
        """

    @feature(fails=True)
    def test_input_values_fail(self):
        """
        When I visit test page "basic_page"
        And I fill in "username" with "Danni"
        Then input "username" has value "Ricky"
        """

    # Chrome's date fields expect input in a localized format, for example,
    # mm/dd/yyyy for en_US (note day and month swapped vis-a-vis ISO format).
    # The test browser locale is set to en_US.

    @skip_if_browser(
        ['firefox', 'phantomjs'],
        "Only Chrome's date fields are in a localized format."
    )
    @feature()
    def test_date_input_localized(self):
        """
        When I visit test page "basic_page"
        And I fill in "dob" with "02141992"
        Then input "dob" has value "1992-02-14"
        """

    @skip_if_browser(
        'chrome',
        "Chrome's date fields are in a localized format."
    )
    @feature()
    def test_date_input(self):
        """
        When I visit test page "basic_page"
        And I fill in "dob" with "1992-02-14"
        Then input "dob" has value "1992-02-14"
        """

    @feature()
    def test_page_title(self):
        """
        When I visit test page "basic_page"
        Then the page title should be "A Basic Page"
        """

    @feature()
    def test_press_of_input_button(self):
        """
        Given I visit test page "basic_page"
        Then I should not see 'You pressed an input button'
        When I press "Reveal"
        Then I should see 'You pressed an input button'
        """

    @feature()
    def test_submit_only(self):
        """
        When I visit test page "basic_page"
        And I submit the only form
        Then the browser's URL should contain "bio="
        And the browser's URL should contain "user="
        """

    @feature()
    def test_submit_action(self):
        """
        When I visit test page "basic_page"
        And I submit the form with action "basic_page.html"
        Then the browser's URL should contain "bio="
        And the browser's URL should contain "user="
        """

    @feature()
    def test_submit_id(self):
        """
        When I visit test page "basic_page"
        And I submit the form with id "the-form"
        Then the browser's URL should contain "bio="
        And the browser's URL should contain "user="
        """

    @feature()
    def test_switch_frame_by_id(self):
        """
        When I visit test page "frame_page"
        Then I should see "This is the main page content"
        And I should not see "This is the frame content"
        When I switch to the frame with id "frame_id"
        Then I should not see "This is the main page content"
        And I should see "This is the frame content"
        When I switch back to the main view
        Then I should see "This is the main page content"
        And I should not see "This is the frame content"
        """

    @feature()
    def test_switch_frame_by_class(self):
        """
        When I visit test page "frame_page"
        Then I should see "This is the main page content"
        And I should not see "This is the frame content"
        When I switch to the frame with class "frame_class"
        Then I should not see "This is the main page content"
        And I should see "This is the frame content"
        When I switch back to the main view
        Then I should see "This is the main page content"
        And I should not see "This is the frame content"
        """

    @feature()
    def test_delayed_id(self):
        """
        When I visit test page "basic_page"
        And I press "Start timer"
        Then I should see an element with id of "delayed_p" within 15 seconds
        """

    @feature()
    def test_delayed_text(self):
        """
        When I visit test page "basic_page"
        And I press "Start timer"
        Then I should see "Time passed" within 15 seconds
        """

    @feature(fails=True)
    def test_delayed_id_failure(self):
        """
        When I visit test page "basic_page"
        And I press "Start timer"
        Then I should see an element with id of "delayed_p" within 5 seconds
        """

    @feature(fails=True)
    def test_delayed_text_failure(self):
        """
        When I visit test page "basic_page"
        And I press "Start timer"
        Then I should see "Time passed" within 5 seconds
        """



"""
Hooks to save screenshots and HTML source of the pages when tests fail.

Assumes a browser instance is stored in ``world.browser``.

Whenever a step fails, the screen shot and the HTML source of the page
displayed in the browser are saved to the current directory. The file names
include the feature file name, scenario number and name and, if applicable,
the example number.

Consider the following feature:

.. code-block:: gherkin

    # features/account.feature
    Feature: Account management

        Scenario: Log in
            Given I open the site
            And I enter username and password
            And I press "Log in"
            Then I should see "Logged in"

If there will be no "Logged in" text when expected, screenshot and the page
source will be saved to, respectively::

    failed_features_account_feature_1_Log_in.png
    failed_features_account_feature_1_Log_in.html

To change the directory where the screenshots are saved, override the constant
``DIRECTORY`` as follows:

.. code-block:: python

    from aloe_webdriver import screenshot_failed

    screenshot_failed.DIRECTORY = '/alternative/directory'

Note that the given directory should already exist.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin,unused-wildcard-import,wildcard-import
from builtins import *
# pylint:enable=redefined-builtin,unused-wildcard-import,wildcard-import

import re
import os

from aloe import after, world

# Pylint cannot infer the attributes on world
# pylint:disable=no-member


DIRECTORY = ''
FORMAT = 'failed_{feature_file}_{scenario_index}_{scenario_name}{outline_index}'


@after.each_step
def take_screenshot(self):
    """Take a screenshot after a failed step."""

    if not self.failed:
        return

    browser = getattr(world, 'browser', None)
    if not browser:
        return

    try:
        scenario_name = self.scenario.name
        scenario_index = \
            self.scenario.feature.scenarios.index(self.scenario) + 1
    except AttributeError:
        scenario_name = self.background.keyword
        scenario_index = 0

    if self.outline is None:
        outline_index_str = ''
    else:
        outline_index = self.scenario.outlines.index(self.outline) + 1
        outline_index_str = '_{}'.format(outline_index)

    base_name = FORMAT.format(
        feature_file=os.path.relpath(self.feature.filename),
        scenario_index=scenario_index,
        scenario_name=scenario_name,
        outline_index=outline_index_str,
    )
    base_name = re.sub(r'\W', '_', base_name, flags=re.UNICODE)
    base_name = os.path.join(DIRECTORY, base_name)

    world.browser.save_screenshot('{}.png'.format(base_name))

    with open('{}.html'.format(base_name), 'w') as page_source_file:
        page_source_file.write(world.browser.page_source)



"""
Base functions for tests.
"""

import os
import socketserver
import threading
import unittest
from contextlib import contextmanager
from functools import wraps
from http.server import SimpleHTTPRequestHandler
from time import sleep

from selenium import webdriver

from aloe.testing import in_directory


def feature(fails=False):
    """
    Decorate a test method to test the feature contained in its docstring.

    For example:
        @feature(failed=False)
        def test_some_feature(self):
            '''
            When I ...
            Then I ...
            '''

    The method code is ignored.
    """

    def outer(func):
        """
        A decorator to run the function as the feature contained in docstring.
        """

        @wraps(func)
        @in_directory(os.path.dirname(__file__))
        def inner(self):
            """Run the scenario from docstring."""

            scenario = func.__doc__

            # Make it possible to reference SERVER_HOST in URLs inside
            # scenarios
            scenario = scenario.replace(
                'SERVER_HOST',
                os.environ.get('SERVER_HOST', '0.0.0.0')
            )

            feature_string = """
            Feature: {name}
            Scenario: {name}
            {scenario_string}
            """.format(name=func.__name__, scenario_string=scenario)

            result = self.run_feature_string(feature_string)

            if fails:
                self.assertFalse(result.success)
            else:
                self.assertTrue(result.success)

        return inner

    return outer


class TestRequestHandler(SimpleHTTPRequestHandler):
    """A handler serving the test pages."""

    def translate_path(self, path):
        """Serve the pages directory instead of the current directory."""

        pages_dir = os.path.relpath(
            os.path.join(os.path.dirname(__file__), 'html_pages'))

        return SimpleHTTPRequestHandler.translate_path(
            self, '/' + pages_dir + path)

    def do_GET(self):
        """
        Artificially slow down the response to make sure there are no race
        conditions.
        """

        sleep(0.5)

        return SimpleHTTPRequestHandler.do_GET(self)

    def log_message(self, *args, **kwargs):
        """Turn off logging."""
        pass


class TestServer(socketserver.TCPServer):
    """Server for the test pages."""

    allow_reuse_address = True

    def get_request(self):
        """Set a timeout on the request socket."""

        request, addr = socketserver.TCPServer.get_request(self)
        request.settimeout(2)  # pylint:disable=no-member
        return request, addr


@contextmanager
def test_server():
    """A context manager starting a server for the test pages."""

    port = 7755

    server = TestServer(('', port), TestRequestHandler)
    server_thread = threading.Thread(target=server.serve_forever)
    server_thread.start()

    # When running the browser in Docker, pass the host address
    # to allow the container to access the server on the host
    if 'SERVER_HOST' in os.environ:
        address = (os.environ['SERVER_HOST'], port)
    else:
        address = server.server_address

    yield server, address

    server.shutdown()
    server_thread.join()
    server.server_close()


def browser_type():
    """Browser type selected for the tests."""

    return os.environ.get('BROWSER_TYPE', 'firefox')


def skip_if_browser(browsers, message):
    """Decorator to skip a test with a particular browser type."""

    if not isinstance(browsers, (list, tuple)):
        browsers = [browsers]

    if browser_type() in browsers:
        return unittest.skip(message)
    return lambda func: func


def create_browser():
    """Create a Selenium browser for tests."""

    if 'SELENIUM_ADDRESS' in os.environ:
        address = 'http://{}/wd/hub'.format(os.environ['SELENIUM_ADDRESS'])

        capabilities = {
            'chrome': webdriver.DesiredCapabilities.CHROME,
            'firefox': webdriver.DesiredCapabilities.FIREFOX,
            'phantomjs': webdriver.DesiredCapabilities.PHANTOMJS,
        }
        try:
            browser = capabilities[browser_type()]
        except KeyError:
            raise ValueError("Invalid BROWSER_TYPE.")

        return webdriver.Remote(
            address,
            desired_capabilities=browser,
        )

    browsers = {
        'chrome': webdriver.Chrome,
        'firefox': webdriver.Firefox,
        'phantomjs': webdriver.PhantomJS,
    }
    driver = browsers[browser_type()]

    # Explicitly specify the browser locale for the date input tests to work
    # regardless of the user's settings
    old_lc_all = os.environ.get('LC_ALL', '')
    try:
        os.environ['LC_ALL'] = 'en_US'
        return driver()
    finally:
        os.environ['LC_ALL'] = old_lc_all



"""
Steps for selecting elements using CSS selectors.

Like with steps based on HTML id, these steps should be used cautiously to
avoid creating tests that do not describe the behaviours of your application.
See :ref:`good-bdd`.

.. note::

    Be aware these steps require jQuery_. If jQuery_ is not present it will be
    added (v1.12).

.. _jQuery: https://jquery.com/
"""

from functools import wraps
from time import sleep

from aloe import step
from aloe import world

from aloe_webdriver.util import (
    wait_for,
)

from selenium.common.exceptions import WebDriverException

# pylint:disable=missing-docstring

# Pylint cannot infer the attributes on world
# pylint:disable=no-member


def load_script(browser, url):
    """Ensure that JavaScript at a given URL is available to the browser."""
    if browser.current_url.startswith('file:'):
        url = 'https:' + url
    browser.execute_script("""
    var script_tag = document.createElement("script");
    script_tag.setAttribute("type", "text/javascript");
    script_tag.setAttribute("src", arguments[0]);
    document.getElementsByTagName("head")[0].appendChild(script_tag);
    """, url)


JQUERY = '//ajax.googleapis.com/ajax/libs/jquery/1.12.0/jquery.min.js'


def is_jquery_not_defined_error(msg):
    """
    Check whether the JavaScript error message is due to jQuery not
    being available.
    """

    # Firefox: '$ is not defined'
    # Chrome: 'unknown error: $ is not defined'
    # PhantomJS: JSON with 'Can't find variable: $'
    return any(txt in msg for txt in (
        "$ is not defined",
        "Can't find variable: $",
    ))


def load_jquery(func):
    """
    A decorator to ensure a function is run with jQuery available.

    If an exception from a function indicates jQuery is missing, it is loaded
    and the function re-executed.

    The browser to load jQuery into must be the first argument of the function.
    """

    @wraps(func)
    def wrapped(browser, *args, **kwargs):
        """Run the function, loading jQuery if needed."""

        try:
            return func(browser, *args, **kwargs)
        except WebDriverException as ex:
            if not is_jquery_not_defined_error(ex.msg):
                raise

            load_script(browser, JQUERY)

            @wait_for
            def jquery_available():
                """Assert that jQuery has loaded."""
                try:
                    return browser.execute_script('return $')
                except WebDriverException:
                    raise AssertionError("jQuery is not loaded")

            jquery_available()

            return func(browser, *args, **kwargs)

    return wrapped


@load_jquery
def find_elements_by_jquery(browser, selector):
    """Find HTML elements using jQuery-style selectors.

    Ensures that jQuery is available to the browser."""

    return browser.execute_script(
        """return ($ || jQuery)(arguments[0]).get();""", selector)


def find_element_by_jquery(browser, selector):
    """Find a single HTML element using jQuery-style selectors."""
    elements = find_elements_by_jquery(browser, selector)
    if not elements:
        raise AssertionError("No matching element found.")
    if len(elements) > 1:
        raise AssertionError("Multiple matching elements found.")
    return elements[0]


@load_jquery
def find_parents_by_jquery(browser, selector):
    """Find HTML elements' parents using jQuery-style selectors.

    Ensures that jQuery is available to the browser."""

    return browser.execute_script(
        """return ($ || jQuery)(arguments[0]).parent().get();""", selector)


@step(r'There should be an element matching \$\("(.*?)"\)$')
@wait_for
def check_element_by_selector(self, selector):
    """Assert an element exists matching the given selector."""
    elems = find_elements_by_jquery(world.browser, selector)
    if not elems:
        raise AssertionError("Expected matching elements, none found.")


@step(r'There should not be an element matching \$\("(.*?)"\)$')
@wait_for
def check_no_element_by_selector(self, selector):
    """Assert an element does not exist matching the given selector."""
    elems = find_elements_by_jquery(world.browser, selector)
    if elems:
        raise AssertionError("Expected no matching elements, found {}.".format(
            len(elems)))


@step(r'There should be an element matching \$\("(.*?)"\) '
      r'within (\d+) seconds?$')
def wait_for_element_by_selector(self, selector, seconds):
    """
    Assert an element exists matching the given selector within the given time
    period.
    """

    def assert_element_present():
        """Assert an element matching the given selector exists."""
        if not find_elements_by_jquery(world.browser, selector):
            raise AssertionError("Expected a matching element.")

    wait_for(assert_element_present)(timeout=int(seconds))


@step(r'There should be exactly (\d+) elements matching \$\("(.*?)"\)$')
@wait_for
def count_elements_exactly_by_selector(self, number, selector):
    """
    Assert n elements exist matching the given selector.
    """
    elems = find_elements_by_jquery(world.browser, selector)
    number = int(number)
    if len(elems) != number:
        raise AssertionError("Expected {} elements, found {}".format(
            number, len(elems)))


@step(r'I fill in \$\("(.*?)"\) with "(.*?)"$')
@wait_for
def fill_in_by_selector(self, selector, value):
    """Fill in the form element matching the CSS selector."""
    elem = find_element_by_jquery(world.browser, selector)
    elem.clear()
    elem.send_keys(value)


@step(r'I submit \$\("(.*?)"\)')
@wait_for
def submit_by_selector(self, selector):
    """Submit the form matching the CSS selector."""
    elem = find_element_by_jquery(world.browser, selector)
    elem.submit()


@step(r'I check \$\("(.*?)"\)$')
@wait_for
def check_by_selector(self, selector):
    """Check the checkbox matching the CSS selector."""
    elem = find_element_by_jquery(world.browser, selector)
    if not elem.is_selected():
        elem.click()


@step(r'I click \$\("(.*?)"\)$')
@wait_for
def click_by_selector(self, selector):
    """Click the element matching the CSS selector."""
    # No need for separate button press step with selector style.
    elem = find_element_by_jquery(world.browser, selector)
    elem.click()


@step(r'I follow the link \$\("(.*?)"\)$')
@wait_for
def follow_link_by_selector(self, selector):
    """
    Navigate to the href of the element matching the CSS selector.

    N.B. this does not click the link, but changes the browser's URL.
    """
    elem = find_element_by_jquery(world.browser, selector)
    href = elem.get_attribute('href')
    world.browser.get(href)


@step(r'\$\("(.*?)"\) should be selected$')
@wait_for
def is_selected_by_selector(self, selector):
    """Assert the option matching the CSS selector is selected."""
    elem = find_element_by_jquery(world.browser, selector)
    if not elem.is_selected():
        raise AssertionError("Element expected to be selected.")


@step(r'I select \$\("(.*?)"\)$')
@wait_for
def select_by_selector(self, selector):
    """Select the option matching the CSS selector."""
    option = find_element_by_jquery(world.browser, selector)
    selectors = find_parents_by_jquery(world.browser, selector)
    if not selectors:
        raise AssertionError("No parent element found for the option.")
    selector = selectors[0]
    selector.click()
    sleep(0.3)
    option.click()
    if not option.is_selected():
        raise AssertionError(
            "Option should have become selected after clicking it.")



"""
Basic Selenium_ :class:`Webdriver` steps for Aloe_.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import str
# pylint:enable=redefined-builtin

from aloe import step, world

from aloe_webdriver.util import (
    ElementSelector,
    find_any_field,
    find_button,
    find_field,
    find_option,
    option_in_select,
    wait_for,
    string_literal,
)

from selenium.webdriver.support.ui import Select
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.common.keys import Keys
from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
    NoAlertPresentException,
    WebDriverException)

# pylint:disable=missing-docstring

# Pylint cannot infer the attributes on world
# pylint:disable=no-member


def contains_content(browser, content):
    """
    Search for an element that contains the whole of the text we're looking
    for in it or its subelements, but whose children do NOT contain that
    text - otherwise matches <body> or <html> or other similarly useless
    things.
    """
    for elem in browser.find_elements_by_xpath(str(
            '//*[contains(normalize-space(.), {content}) '
            'and not(./*[contains(normalize-space(.), {content})])]'
            .format(content=string_literal(content)))):

        try:
            if elem.is_displayed():
                return True
        except StaleElementReferenceException:
            pass

    return False


# Navigation ################################################################


@step('I visit "(.*?)"$')
@step('I go to "(.*?)"$')
def visit(self, url):
    """Navigate to the provided (fully qualified) URL."""
    world.browser.get(url)


@step('The browser\'s URL should be "([^"]*)"$')
@step('I should be at "([^"]*)"$')
@wait_for
def url_should_be(self, url):
    """Assert the absolute URL of the browser is as provided."""

    if world.browser.current_url != url:
        raise AssertionError(
            "Browser URL expected to be {!r}, got {!r}.".format(
                url, world.browser.current_url))


@step('''The browser's URL should contain "([^"]*)"$''')
@wait_for
def url_should_contain(self, url):
    """Assert the absolute URL of the browser contains the provided."""

    if url not in world.browser.current_url:
        raise AssertionError(
            "Browser URL expected to contain {!r}, got {!r}.".format(
                url, world.browser.current_url))


@step('''The browser's URL should not contain "([^"]*)"$''')
@wait_for
def url_should_not_contain(self, url):
    """Assert the absolute URL of the browser does not contain the provided."""

    if url in world.browser.current_url:
        raise AssertionError(
            "Browser URL expected not to contain {!r}, got {!r}.".format(
                url, world.browser.current_url))


@step(r'The page title should be "([^"]*)"')
@step(r"The page title should be '([^']*)'")
@wait_for
def page_title(self, title):
    """
    Assert the page title matches the given text.
    """

    if world.browser.title != title:
        raise AssertionError(
            "Page title expected to be {!r}, got {!r}.".format(
                title, world.browser.title))

# Links #####################################################################


@step('I click "([^"]*)"$')
@step("I click '([^']*)'$")
@wait_for
def click(self, name):
    """Click the link with the provided link text."""
    try:
        elem = world.browser.find_element_by_link_text(name)
    except NoSuchElementException:
        raise AssertionError(
            "Cannot find the link with text '{}'.".format(name))
    elem.click()


@step('I should see a link with the url "([^"]*)"$')
@wait_for
def should_see_link(self, link_url):
    """Assert a link with the provided URL is visible on the page."""

    elements = ElementSelector(
        world.browser,
        str('//a[@href="%s"]' % link_url),
        filter_displayed=True,
    )
    if not elements:
        raise AssertionError("Expected link not found.")


@step('I should see a link to "([^"]*)" with the url "([^"]*)"$')
@step("I should see a link to '([^']*)' with the url '([^']*)'$")
@wait_for
def should_see_link_text(self, link_text, link_url):
    """Assert a link with the provided text points to the provided URL."""

    elements = ElementSelector(
        world.browser,
        str('//a[@href="%s"][./text()="%s"]' % (link_url, link_text)),
        filter_displayed=True,
    )
    if not elements:
        raise AssertionError("Expected link not found.")


@step('I should see a link that contains the text "([^"]*)" and '
      'the url "([^"]*)"$')
@step("I should see a link that contains the text '([^']*)' and "
      "the url '([^']*)'$")
@wait_for
def should_include_link_text(self, link_text, link_url):
    """
    Assert a link containing the provided text points to the provided URL.
    """

    elements = ElementSelector(
        world.browser,
        str('//a[@href="%s"][contains(., %s)]' %
            (link_url, string_literal(link_text))),
        filter_displayed=True,
    )
    if not elements:
        raise AssertionError("Expected link not found.")


# ID based checks ###########################################################


@step('The element with id of "([^"]*)" contains "([^"]*)"$')
@step("The element with id of '([^']*)' contains '([^']*)'$")
@wait_for
def element_contains(self, element_id, value):
    """
    Assert provided content is contained within an element found by ``id``.
    """

    elements = ElementSelector(
        world.browser,
        str('id("{id}")[contains(., "{value}")]'.format(
            id=element_id, value=value)),
        filter_displayed=True,
    )
    if not elements:
        raise AssertionError("Expected element not found.")


@step('The element with id of "([^"]*)" does not contain "([^"]*)"$')
@step("The element with id of '([^']*)' does not contain '([^']*)'$")
@wait_for
def element_not_contains(self, element_id, value):
    """
    Assert provided content is not contained within an element found by ``id``.
    """
    elem = world.browser.find_elements_by_xpath(str(
        'id("{id}")[contains(., "{value}")]'.format(
            id=element_id, value=value)))
    assert not elem, \
        "Expected element not to contain the given text."


@step(r'I should see an element with id of "([^"]*)" within (\d+) seconds?$')
def should_see_id_in_seconds(self, element_id, timeout):
    """
    Assert an element with the given ``id`` is visible within n seconds.
    """

    def check_element():
        """Check for the element with the given id."""

        assert ElementSelector(
            world.browser,
            'id("%s")' % element_id,
            filter_displayed=True,
        ), "Expected element with given id."

    wait_for(check_element)(timeout=int(timeout))


@step('I should see an element with id of "([^"]*)"$')
@wait_for
def should_see_id(self, element_id):
    """
    Assert an element with the given ``id`` is visible.
    """

    elements = ElementSelector(
        world.browser,
        'id("%s")' % element_id,
        filter_displayed=True,
    )
    if not elements:
        raise AssertionError("Expected element with given id.")


@step('I should not see an element with id of "([^"]*)"$')
@wait_for
def should_not_see_id(self, element_id):
    """
    Assert an element with the given ``id`` is not visible.
    """

    elements = ElementSelector(
        world.browser,
        'id("%s")' % element_id,
        filter_displayed=True,
    )
    if elements:
        raise AssertionError("Expected element with given id to be absent.")


@step(r'Element with id "([^"]*)" should be focused')
@wait_for
def element_focused(self, id_):
    """
    Assert the element is focused.
    """

    try:
        elem = world.browser.find_element_by_id(id_)
    except NoSuchElementException:
        raise AssertionError("Element with ID '{}' not found.".format(id_))

    focused = world.browser.switch_to.active_element

    # Elements don't have __ne__ defined, cannot test for inequality
    if not elem == focused:
        raise AssertionError("Expected element to be focused.")


@step(r'Element with id "([^"]*)" should not be focused')
@wait_for
def element_not_focused(self, id_):
    """
    Assert the element is not focused.
    """

    try:
        elem = world.browser.find_element_by_id(id_)
    except NoSuchElementException:
        raise AssertionError("Element with ID '{}' not found.".format(id_))

    focused = world.browser.switch_to.active_element

    # Elements don't have __ne__ defined, cannot test for inequality
    if elem == focused:
        raise AssertionError("Expected element not to be focused.")


# Text ######################################################################


@step(r'I should see "([^"]+)" within (\d+) seconds?$')
@step(r"I should see '([^']+)' within (\d+) seconds?$")
def should_see_in_seconds(self, text, timeout):
    """
    Assert provided text is visible within n seconds.

    Be aware this text could be anywhere on the screen. Also be aware that
    it might cross several HTML nodes. No determination is made between
    block and inline nodes. Whitespace can be affected.
    """

    def check_element():
        """Check for an element with the given content."""

        assert contains_content(world.browser, text), \
            "Expected element with the given text."

    wait_for(check_element)(timeout=int(timeout))


@step('I should see "([^"]+)"$')
@step("I should see '([^']+)'$")
@step('I see "([^"]+)"$')
@step("I see '([^']+)'$")
@wait_for
def should_see(self, text):
    """
    Assert provided text is visible.

    Be aware this text could be anywhere on the screen. Also be aware that
    it might cross several HTML nodes. No determination is made between
    block and inline nodes. Whitespace can be affected.
    """
    if not contains_content(world.browser, text):
        raise AssertionError("Expected content not found.")


@step('I should not see "([^"]+)"$')
@step("I should not see '([^']+)'$")
@wait_for
def should_not_see(self, text):
    """
    Assert provided text is not visible.

    Be aware that because of the caveats of the positive case, the text MAY
    be on the screen in a slightly different form.
    """
    if contains_content(world.browser, text):
        raise AssertionError("Content unexpectedly found.")


# Forms #####################################################################


@step('I should see a form that goes to "([^"]*)"$')
@wait_for
def see_form(self, url):
    """
    Assert the existence of a HTML form that submits to the given URL.
    """

    elements = ElementSelector(
        world.browser,
        str('//form[@action="%s"]' % url),
        filter_displayed=True,
    )
    if not elements:
        raise AssertionError("Expected form not found.")


DATE_FIELDS = (
    'datetime',
    'datetime-local',
    'date',
)


TEXT_FIELDS = (
    'text',
    'textarea',
    'password',
    'month',
    'time',
    'week',
    'number',
    'range',
    'email',
    'url',
    'tel',
    'color',
)


@step('I fill in "([^"]*)" with "([^"]*)"$')
@step("I fill in '([^']*)' with '([^']*)'$")
@wait_for
def fill_in_textfield(self, field_name, value):
    """
    Fill in the HTML input with given label (recommended), name or id with
    the given text.

    Supported input types are text, textarea, password, month, time, week,
    number, range, email, url, tel and color.
    """

    date_field = find_any_field(world.browser,
                                DATE_FIELDS,
                                field_name)
    if date_field:
        field = date_field
    else:
        field = find_any_field(world.browser,
                               TEXT_FIELDS,
                               field_name)

    if not field:
        raise AssertionError(
            "Cannot find a field named '{}'.".format(field_name))

    if date_field:
        field.send_keys(Keys.DELETE)
    else:
        field.clear()

    field.send_keys(value)


@step('I press "([^"]*)"$')
@step("I press '([^']*)'$")
@wait_for
def press_button(self, value):
    """
    Click the button with the given label.
    """
    button = find_button(world.browser, value)
    if not button:
        raise AssertionError(
            "Cannot find a button named '{}'.".format(value))
    button.click()


@step('I click on label "([^"]*)"')
@step("I click on label '([^']*)'")
@wait_for
def click_on_label(self, label):
    """
    Click on the given label.

    On a correctly set up form this will highlight the appropriate field.
    """

    elem = ElementSelector(
        world.browser,
        str('//label[normalize-space(text())=%s]' % string_literal(label)),
        filter_displayed=True,
    )
    if not elem:
        raise AssertionError(
            "Cannot find a label with text '{}'.".format(label))
    elem.click()


@step(r'Input "([^"]*)" (?:has|should have) value "([^"]*)"')
@step(r"Input '([^']*)' (?:has|should have) value '([^']*)'")
@wait_for
def input_has_value(self, field_name, value):
    """
    Assert the form input with label (recommended), name or id has given value.
    """
    text_field = find_any_field(world.browser,
                                DATE_FIELDS + TEXT_FIELDS,
                                field_name)
    if text_field is False:
        raise AssertionError(
            "Can not find a field named {!r}.".format(field_name))

    actual = text_field.get_attribute('value')
    if actual != value:
        raise AssertionError(
            "Field value expected to be {!r}, got {!r}.".format(
                value, actual))


@step(r'I submit the only form')
@wait_for
def submit_the_only_form(self):
    """
    Look for a form on the page and submit it.

    Asserts if more than one form exists.
    """
    form = ElementSelector(world.browser, str('//form'))
    assert form, "Cannot find a form on the page."
    form.submit()


@step(r'I submit the form with id "([^"]*)"')
@wait_for
def submit_form_id(self, id_):
    """
    Submit the form with given id (used to disambiguate between multiple
    forms).
    """
    form = ElementSelector(
        world.browser,
        str('id("{id}")'.format(id=id_)),
    )
    assert form, "Cannot find a form with ID '{}' on the page.".format(id_)
    form.submit()


@step(r'I submit the form with action "([^"]*)"')
def submit_form_action(self, url):
    """
    Submit the form with the given action URL (i.e. the form that submits to
    ``/post/my/data``).
    """
    form = ElementSelector(
        world.browser,
        str('//form[@action="%s"]' % url),
    )
    assert form, \
        "Cannot find a form with action '{}' on the page.".format(url)
    form.submit()


# Checkboxes ################################################################


@step('I check "([^"]*)"$')
@step("I check '([^']*)'$")
@wait_for
def check_checkbox(self, value):
    """Check the checkbox with label (recommended), name or id."""
    check_box = find_field(world.browser, 'checkbox', value)
    assert check_box, "Cannot find checkbox '{}'.".format(value)
    if not check_box.is_selected():
        check_box.click()


@step('I uncheck "([^"]*)"$')
@step("I uncheck '([^']*)'$")
@wait_for
def uncheck_checkbox(self, value):
    """Uncheck the checkbox with label (recommended), name or id."""
    check_box = find_field(world.browser, 'checkbox', value)
    assert check_box, "Cannot find checkbox '{}'.".format(value)
    if check_box.is_selected():
        check_box.click()


@step('The "([^"]*)" checkbox should be checked$')
@step("The '([^']*)' checkbox should be checked$")
@wait_for
def assert_checked_checkbox(self, value):
    """Assert the checkbox with label (recommended), name or id is checked."""
    check_box = find_field(world.browser, 'checkbox', value)
    assert check_box, "Cannot find checkbox '{}'.".format(value)
    assert check_box.is_selected(), "Check box should be selected."


@step('The "([^"]*)" checkbox should not be checked$')
@step("The '([^']*)' checkbox should not be checked$")
@wait_for
def assert_not_checked_checkbox(self, value):
    """
    Assert the checkbox with label (recommended), name or id is not checked.
    """
    check_box = find_field(world.browser, 'checkbox', value)
    assert check_box, "Cannot find checkbox '{}'.".format(value)
    assert not check_box.is_selected(), "Check box should not be selected."


# Selects ###################################################################


@step('I select "([^"]*)" from "([^"]*)"$')
@step("I select '([^']*)' from '([^']*)'$")
@wait_for
def select_single_item(self, option_name, select_name):
    """
    Select the named option from select with label (recommended), name or id.
    """
    option_box = find_option(world.browser, select_name, option_name)
    assert option_box, "Cannot find option '{}'.".format(option_name)
    option_box.click()


@step('I select the following from "([^"]*?)":?$')
@step("I select the following from '([^']*?)':?$")
@wait_for
def select_multi_items(self, select_name):
    """
    Select multiple options from select with label (recommended), name, or
    id. Pass a multiline string of options. e.g.

    .. code-block:: gherkin

        When I select the following from "Contact Methods":
            \"\"\"
            Email
            Phone
            Fax
            \"\"\"
    """
    # Ensure only the options selected are actually selected
    option_names = self.multiline.split('\n')
    select_box = find_field(world.browser, 'select', select_name)
    assert select_box, "Cannot find a '{}' select.".format(select_name)

    select = Select(select_box)
    select.deselect_all()

    for option in option_names:
        try:
            select.select_by_value(option)
        except NoSuchElementException:
            try:
                select.select_by_visible_text(option)
            except NoSuchElementException:
                raise AssertionError("Cannot find option: '{}'.".format(option))


@step('The "([^"]*)" option from "([^"]*)" should be selected$')
@step("The '([^']*)' option from '([^']*)' should be selected$")
@wait_for
def assert_single_selected(self, option_name, select_name):
    """
    Assert the given option is selected from the select with label
    (recommended), name or id.

    If multiple selections are supported other options may be selected.
    """
    option = find_option(world.browser, select_name, option_name)
    assert option.is_selected(), "Option should be selected."


@step('The following options from "([^"]*?)" should be selected:?$')
@step("The following options from '([^']*?)' should be selected:?$")
@wait_for
def assert_multi_selected(self, select_name):
    select_box = find_field(world.browser, 'select', select_name)
    assert select_box, "Cannot find a '{}' select.".format(select_name)

    option_names = self.multiline.split('\n')
    option_elems = select_box.find_elements_by_xpath(str('./option'))

    # Check only the options that are specified are selected
    for option in option_elems:
        if option.get_attribute('id') in option_names or \
                option.get_attribute('name') in option_names or \
                option.get_attribute('value') in option_names or \
                option.text in option_names:
            assert option.is_selected(), "Option should be selected."
        else:
            assert not option.is_selected(), \
                "Option should not be selected."


@step(r'I should see option "([^"]*)" in selector "([^"]*)"')
@step(r"I should see option '([^']*)' in selector '([^']*)'")
@wait_for
def select_contains(self, option, id_):
    """Assert the select contains the given option."""
    assert option_in_select(world.browser, id_, option) is not None, \
        "Selector should have the given option."


@step(r'I should not see option "([^"]*)" in selector "([^"]*)"')
@step(r"I should not see option '([^']*)' in selector '([^']*)'")
@wait_for
def select_does_not_contain(self, option, id_):
    """Assert the select does not contain the given option."""
    assert option_in_select(world.browser, id_, option) is None, \
        "Selector should not have the given option."


# Radios ####################################################################


@step('I choose "([^"]*)"$')
@step("I choose '([^']*)'$")
@wait_for
def choose_radio(self, value):
    """
    Click (and choose) the radio button with the given label (recommended),
    name or id.
    """
    box = find_field(world.browser, 'radio', value)
    assert box, "Cannot find a '{}' radio button.".format(value)
    box.click()


@step('The "([^"]*)" option should be chosen$')
@step("The '([^']*)' option should be chosen$")
@wait_for
def assert_radio_selected(self, value):
    """
    Assert the radio button with the given label (recommended), name or id is
    chosen.
    """
    radio = find_field(world.browser, 'radio', value)
    assert radio, "Cannot find a '{}' radio button.".format(value)
    assert radio.is_selected(), "Radio button should be selected."


@step('The "([^"]*)" option should not be chosen$')
@step("The '([^']*)' option should not be chosen$")
@wait_for
def assert_radio_not_selected(self, value):
    """
    Assert the radio button with the given label (recommended), name or id is
    not chosen.
    """
    radio = find_field(world.browser, 'radio', value)
    assert radio, "Cannot find a '{}' radio button.".format(value)
    assert not radio.is_selected(), "Radio button should not be selected."


# Alerts ####################################################################


@step('I accept the alert')
def accept_alert(self):
    """
    Accept the alert.
    """

    try:
        alert = Alert(world.browser)
        alert.accept()
    except WebDriverException:
        # PhantomJS is kinda poor
        pass


@step('I dismiss the alert')
def dismiss_alert(self):
    """
    Dismiss the alert.
    """

    try:
        alert = Alert(world.browser)
        alert.dismiss()
    except WebDriverException:
        # PhantomJS is kinda poor
        pass


@step(r'I should see an alert with text "([^"]*)"')
@step(r"I should see an alert with text '([^']*)'")
def check_alert(self, text):
    """
    Assert an alert is showing with the given text.
    """

    try:
        alert = Alert(world.browser)
        if alert.text != text:
            raise AssertionError(
                "Alert text expected to be {!r}, got {!r}.".format(
                    text, alert.text))
    except WebDriverException:
        # PhantomJS is kinda poor
        pass


@step('I should not see an alert')
def check_no_alert(self):
    """
    Assert there is no alert.
    """

    try:
        alert = Alert(world.browser)
        raise AssertionError("Should not see an alert. Alert '%s' shown." %
                             alert.text)
    except NoAlertPresentException:
        pass

# Tooltips ##################################################################


def find_by_tooltip(browser, tooltip):
    """
    Find elements with the given tooltip.

    :param browser: ``world.browser``
    :param tooltip: Tooltip to search for

    Returns: an :class:`ElementSelector`
    """

    return ElementSelector(
        world.browser,
        str('//*[@title=%(tooltip)s or @data-original-title=%(tooltip)s]' %
            dict(tooltip=string_literal(tooltip))),
        filter_displayed=True,
    )


@step(r'I should see an element with tooltip "([^"]*)"')
@step(r"I should see an element with tooltip '([^']*)'")
@wait_for
def see_tooltip(self, tooltip):
    """
    Assert an element with the given tooltip (title) is visible.

    N.B. tooltip may not be visible.
    """

    assert find_by_tooltip(world.browser, tooltip), \
        "Expected element with given tooltip."


@step(r'I should not see an element with tooltip "([^"]*)"')
@step(r"I should not see an element with tooltip '([^']*)'")
@wait_for
def no_see_tooltip(self, tooltip):
    """
    Assert an element with the given tooltip (title) is not visible.
    """

    assert not find_by_tooltip(world.browser, tooltip), \
        "Expected no elements with given tooltip."


@step(r'I (?:click|press) the element with tooltip "([^"]*)"')
@step(r"I (?:click|press) the element with tooltip '([^']*)'")
def press_by_tooltip(self, tooltip):
    """
    Click on a HTML element with a given tooltip.

    This is very useful if you're clicking on icon buttons, etc.
    """
    for button in find_by_tooltip(world.browser, tooltip):
        try:
            button.click()
            break
        except:  # pylint:disable=bare-except
            pass
    else:
        raise AssertionError("No button with tooltip '{0}' found"
                             .format(tooltip))

# Frames ####################################################################


@step(r'I switch to the frame with id "([^"]*)"')
def switch_to_frame_with_id(self, frame):
    """Swap Selenium's context to the given frame or iframe."""
    elem = world.browser.find_element_by_id(frame)
    world.browser.switch_to.frame(elem)


@step(r'I switch to the frame with class "([^"]*)"')
def switch_to_frame_with_class(self, frame):
    """Swap Selenium's context to the given frame or iframe."""
    elem = world.browser.find_element_by_class_name(frame)
    world.browser.switch_to.frame(elem)


@step(r'I switch back to the main view')
def switch_to_main(self):
    """Swap Selenium's context back to the main window."""
    world.browser.switch_to.default_content()



"""
Steps for testing.
"""

import os
from contextlib import contextmanager

try:
    reload
except NameError:
    # pylint:disable=no-name-in-module,redefined-builtin
    from importlib import reload
    # pylint:enable=no-name-in-module,redefined-builtin

from aloe import around, before, step, world

import aloe_webdriver
import aloe_webdriver.css

from aloe_webdriver.tests.base import create_browser, test_server

# This module is reloaded during testing in order to re-register the steps and
# callbacks. Make sure the modules where the steps are defined are, too.
reload(aloe_webdriver)
reload(aloe_webdriver.css)

if os.environ.get('TAKE_SCREENSHOTS'):
    # Only register the screenshot steps if asked to
    import aloe_webdriver.screenshot_failed
    reload(aloe_webdriver.screenshot_failed)

    SCREENSHOTS_DIR = os.environ.get('SCREENSHOTS_DIR')
    if SCREENSHOTS_DIR:
        aloe_webdriver.screenshot_failed.DIRECTORY = SCREENSHOTS_DIR


@around.all
@contextmanager
def with_browser():
    """Start a browser for the tests."""
    world.browser = create_browser()

    yield

    world.browser.quit()
    delattr(world, 'browser')


@around.all
@contextmanager
def with_test_server():
    """Start a server for the test pages."""

    with test_server() as (server, address):
        world.server = server
        world.base_address = address
        yield


@before.each_feature
def reset_page(feature):
    """Reset the browser before each feature."""
    world.browser.get('about:blank')


@step(r'I visit test page "([^"]+)"')
def visit_test_page(self, page):
    """Open a test page in the browser."""
    self.when('I visit "http://{address[0]}:{address[1]}/{page}.html"'.format(
        address=world.base_address,
        page=page,
    ))



"""
Test a variety of different buttons.
"""

from aloe.testing import FeatureTest

from aloe_webdriver.tests.base import feature

# pylint:disable=line-too-long


class TestSteps(FeatureTest):
    """Test steps."""

    @feature()
    def test_press_of_submit_button_by_name(self):
        """
        Given I visit test page "button_page"
        When I press "submit_button"
        Then I should see "You pressed the submit button"
        """

    @feature()
    def test_press_of_submit_button_by_value(self):
        """
        Given I visit test page "button_page"
        When I press "Submit button"
        Then I should see "You pressed the submit button"
        """

    @feature()
    def test_press_of_reset_button_by_name(self):
        """
        Given I visit test page "button_page"
        When I press "reset_button"
        Then I should see "You pressed the reset button"
        """

    @feature()
    def test_press_of_reset_button_by_value(self):
        """
        Given I visit test page "button_page"
        When I press "Reset button"
        Then I should see "You pressed the reset button"
        """

    @feature()
    def test_press_of_input_button_by_name(self):
        """
        Given I visit test page "button_page"
        When I press "input_button"
        Then I should see "You pressed the input button"
        """

    @feature()
    def test_press_of_input_button_by_value(self):
        """
        Given I visit test page "button_page"
        When I press "Input button"
        Then I should see "You pressed the input button"
        """

    @feature()
    def test_press_of_image_button_by_name(self):
        """
        Given I visit test page "button_page"
        When I press "image_button"
        Then I should see "You pressed the image button"
        """

    @feature()
    def test_press_of_button_element_by_name(self):
        """
        Given I visit test page "button_page"
        When I press "button_element"
        Then I should see "You pressed the button element"
        """

    @feature()
    def test_press_of_button_element_by_value(self):
        """
        Given I visit test page "button_page"
        When I press "Button element"
        Then I should see "You pressed the button element"
        """

    @feature()
    def test_press_of_anchor_button_by_name(self):
        """
        Given I visit test page "button_page"
        When I press "anchor_button"
        Then I should see "You pressed the anchor button"
        """

    @feature()
    def test_press_of_anchor_button_by_value(self):
        """
        Given I visit test page "button_page"
        When I press "Anchor button"
        Then I should see "You pressed the anchor button"
        """

    @feature()
    def test_press_of_div_button_by_name(self):
        """
        Given I visit test page "button_page"
        When I press "div_button"
        Then I should see "You pressed the div button"
        """

    @feature()
    def test_press_of_div_button_by_value(self):
        """
        Given I visit test page "button_page"
        When I press "Div button"
        Then I should see "You pressed the div button"
        """

    @feature()
    def test_press_of_span_button_by_name(self):
        """
        Given I visit test page "button_page"
        When I press "span_button"
        Then I should see "You pressed the span button"
        """

    @feature()
    def test_press_of_span_button_by_value(self):
        """
        Given I visit test page "button_page"
        When I press "Span button"
        Then I should see "You pressed the span button"
        """

    @feature()
    def test_press_of_paragraph_button_by_name(self):
        """
        Given I visit test page "button_page"
        When I press "paragraph_button"
        Then I should see "You pressed the paragraph button"
        """

    @feature()
    def test_press_of_paragraph_button_by_value(self):
        """
        Given I visit test page "button_page"
        When I press "Paragraph button"
        Then I should see "You pressed the paragraph button"
        """



"""
Django-specific extensions for use with aloe_django_.

.. _aloe_django: https://github.com/aloetesting/aloe_django
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin,unused-wildcard-import,wildcard-import
from builtins import *
# pylint:enable=redefined-builtin,unused-wildcard-import,wildcard-import

try:
    from urllib.parse import urljoin  # pylint:disable=import-error
except ImportError:
    from urlparse import urljoin  # pylint:disable=import-error

from aloe import step
from aloe_django import django_url

# make sure the steps are loaded
import aloe_webdriver  # pylint:disable=unused-import


@step(r'I visit site page "([^"]*)"')
def visit_page(self, page):
    """
    Visit the specific page of the site, e.g.

    .. code-block:: gherkin

        When I visit site page "/users"
    """

    url = urljoin(django_url(self), page)
    self.given('I visit "%s"' % url)



# -*- coding: utf-8 -*-
"""
Test saving screenshots after failed steps.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin,unused-wildcard-import,wildcard-import
from builtins import *
# pylint:enable=redefined-builtin,unused-wildcard-import,wildcard-import

import re
import os
import shutil
import tempfile
from glob import iglob

from aloe.testing import FeatureTest, in_directory


@in_directory(os.path.dirname(__file__))
class TestScreenshots(FeatureTest):
    """Test saving screenshots after failed steps."""

    dir_path = None

    def cleanup_screenshots(self):
        """Clean up any screenshots taken."""

        for ext in ('png', 'html'):
            for filename in iglob('failed_*.{}'.format(ext)):
                os.unlink(filename)

        # Delete temporary directory if any.
        if self.dir_path:
            shutil.rmtree(self.dir_path)

    def setUp(self):
        """Enable the hooks for taking screenshots."""

        super(TestScreenshots, self).setUp()

        self.cleanup_screenshots()

        # This environment variable controls whether the screenshot hooks are
        # registered in tests/features/steps.py
        os.environ['TAKE_SCREENSHOTS'] = '1'

    def tearDown(self):
        """Remove all the screenshot files."""

        del os.environ['TAKE_SCREENSHOTS']

        if 'SCREENSHOTS_DIR' in os.environ:
            del os.environ['SCREENSHOTS_DIR']

        self.cleanup_screenshots()

        super(TestScreenshots, self).tearDown()

    def feature_name(self, test_result):
        """
        The feature file name as visible in the failed screenshot/page source
        file name.

        Needed because the feature will be put in a temporary file with a
        different name each time.

        :param test_result: Result of running the feature
        """

        feature_filename = os.path.relpath(test_result.tests_run[0])
        return re.sub(r'\W', '_', feature_filename)

    def show_files(self):
        """
        Print all the file names matching the screenshot/page source pattern in
        the directory.
        """

        for filename in iglob('failed_*'):
            print(filename)

        # List content of temporary directory, if defined.
        if self.dir_path:
            for filename in iglob(os.path.join(self.dir_path, '*')):
                print(filename)

    def assert_file_present(self, filename, message):
        """Assert a file exists."""

        if os.path.exists(filename):
            return

        self.show_files()
        raise AssertionError(message)

    def assert_file_absent(self, filename, message):
        """Assert a file does not exist."""

        if not os.path.exists(filename):
            return

        self.show_files()
        raise AssertionError(message)

    def test_failed_screenshots(self):
        """Test that failed tests screenshots and page source are recorded."""

        feature_string = """
Feature: Test screenshots on failed steps

Scenario: This scenario succeeds
    When I visit test page "basic_page"
    Then I should see "Hello there"

Scenario: This scenario fails
    When I visit test page "basic_page"
    Then I should see "A unicorn"
"""

        result = self.run_feature_string(feature_string)
        feature = self.feature_name(result)

        self.assert_file_absent(
            'failed_{}_1_This_scenario_succeeds.png'.format(feature),
            "Successful scenario should not be screenshotted."
        )
        self.assert_file_absent(
            'failed_{}_1_This_scenario_succeeds.html'.format(feature),
            "Successful scenario page source should not be saved."
        )

        self.assert_file_present(
            'failed_{}_2_This_scenario_fails.png'.format(feature),
            "Failed scenario should be screenshotted."
        )
        with open('failed_{}_2_This_scenario_fails.html'.format(feature)) \
                as page_source:
            self.assertIn("<title>A Basic Page</title>", page_source.read(),
                          "Failed scenario page source should be saved.")

    def test_failed_background(self):
        """Test that failure of a background step is recorded."""

        feature_string = """
# language: zh-CN
功能: 背景失败

    背景:
        当I visit test page "basic_page"
        那么I should see "A unicorn"

    场景: 必须有场景
        那么I should see "Hello there"
"""

        result = self.run_feature_string(feature_string)
        feature = self.feature_name(result)

        self.assert_file_present(
            'failed_{}_0_背景.png'.format(feature),
            "Failed background should be screenshotted."
        )
        with open('failed_{}_0_背景.html'.format(feature)) as page_source:
            self.assertIn("<title>A Basic Page</title>", page_source.read(),
                          "Failed background page source should be saved.")

    def test_failed_examples(self):
        """Test that failure in an example is recorded."""

        feature_string = """
Feature: Test screenshots on examples

Scenario Outline: Succeeds sometimes
    When I visit test page "basic_page"
    Then I should see "<text>"

    Examples:
        | text        |
        | Hello there |
        | A unicorn   |
"""

        result = self.run_feature_string(feature_string)
        feature = self.feature_name(result)

        self.assert_file_absent(
            'failed_{}_1_Succeeds_sometimes_1.png'.format(feature),
            "Successful example should not be screenshotted."
        )
        self.assert_file_present(
            'failed_{}_1_Succeeds_sometimes_2.png'.format(feature),
            "Failed example should be screenshotted."
        )
        with open('failed_{}_1_Succeeds_sometimes_2.html'.format(feature)) \
                as page_source:
            self.assertIn("<title>A Basic Page</title>", page_source.read(),
                          "Failed example page source should be saved.")

    def test_failed_screenshots_to_dir(self):
        """
        Test failed tests screenshots are saved in the specified directory.
        """

        feature_string = """
Feature: Test screenshots on failed steps

Scenario: This scenario fails
    When I visit test page "basic_page"
    Then I should see "A unicorn"
"""
        self.dir_path = tempfile.mkdtemp()
        os.environ['SCREENSHOTS_DIR'] = self.dir_path
        result = self.run_feature_string(feature_string)
        feature = self.feature_name(result)

        template = 'failed_{}_1_This_scenario_fails.{}'
        png_file = os.path.join(self.dir_path, template.format(feature, 'png'))
        html_file = os.path.join(
            self.dir_path, template.format(feature, 'html'))

        self.assert_file_present(
            png_file,
            "Failed scenario should be screenshotted."
        )
        with open(html_file) as page_source:
            self.assertIn("<title>A Basic Page</title>", page_source.read(),
                          "Failed scenario page source should be saved.")



# -*- coding: utf-8 -*-
"""Test step functions directly."""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import unittest
from time import sleep, time

from aloe import world
from aloe_webdriver.util import (
    find_button,
    find_field,
    find_field_by_id,
    find_field_by_label,
    find_field_by_name,
    option_in_select,
    wait_for,
)

from aloe_webdriver.tests.base import create_browser, test_server

# pylint:disable=missing-docstring


class TestUtil(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        world.browser = create_browser()

    def setUp(self):
        with test_server() as (_, address):
            world.browser.get(
                'http://{address[0]}:{address[1]}/{page}.html'.format(
                    address=address,
                    page='basic_page',
                )
            )

    @classmethod
    def tearDownClass(cls):
        world.browser.quit()
        delattr(world, 'browser')

    def test_find_by_id(self):
        assert find_field_by_id(world.browser, 'password', 'pass')

    def test_find_by_name(self):
        assert find_field_by_name(world.browser, 'submit', 'submit_main')
        assert find_field_by_name(world.browser, 'select', 'car_choice')
        assert find_field_by_name(world.browser, 'textarea', 'bio')

    def test_find_by_label(self):
        field = find_field_by_label(world.browser, 'text', 'Username:')
        self.assertEqual(field.get_attribute('id'), 'username')

    def test_no_label(self):
        assert not find_field_by_label(world.browser, 'text', 'NoSuchLabel')

    def test_find_field(self):
        assert find_field(world.browser, 'text', 'username')
        assert find_field(world.browser, 'text', 'Username:')
        assert find_field(world.browser, 'text', 'user')
        assert find_field(world.browser, 'text', 'ชื่อ:')

    def test_find_button(self):
        assert find_button(world.browser, 'submit_main')
        assert find_button(world.browser, 'Submit!')
        assert find_button(world.browser, 'submit_tentative')
        assert find_button(world.browser, 'Submit as tentative')
        assert find_button(world.browser, 'ส่งฟอร์ม')

    def test_option_in_select(self):
        assert option_in_select(world.browser, 'Favorite Colors:', 'Blue')
        assert option_in_select(world.browser, 'Favorite Colors:', 'ฟ้า')


class TestWaitFor(unittest.TestCase):
    """Test wait_for."""

    def test_wait_for(self):
        """
        Test that wait_for retries on assertion errors.
        """

        start_time = time()

        @wait_for
        def seconds_passed(seconds):
            """
            Test that at least the given number of seconds passed since the
            start of the test.
            """

            assert time() - start_time >= seconds
            return True

        # pylint:disable=unexpected-keyword-arg
        # wait_for decorator parses the argument

        with self.assertRaises(AssertionError):
            seconds_passed(3, timeout=1)

        start_time = time()
        assert seconds_passed(3, timeout=5)

    def test_slow_function(self):
        """
        Test that wait_for still waits for the required number of seconds if the
        wrapped function takes a long time to execute.
        """

        start_time = time()

        @wait_for
        def slow_seconds_passed(seconds):
            """
            Test that at least the given number of seconds passed since the
            start of the test.

            Delay the response to simulate a computationally expensive test.
            """

            result = time() - start_time >= seconds
            sleep(7)
            assert result
            return True

        # pylint:disable=unexpected-keyword-arg
        # wait_for decorator parses the argument

        assert slow_seconds_passed(3, timeout=5)



"""
`Aloe-Webdriver` includes several utilities for writing Selenium_ tests.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import str
# pylint:enable=redefined-builtin

import operator
from copy import copy
from functools import wraps
from time import time, sleep

try:
    reduce
except NameError:
    from functools import reduce  # pylint:disable=redefined-builtin,ungrouped-imports

from selenium.common.exceptions import NoSuchElementException

# pylint:disable=missing-docstring,redefined-outer-name,redefined-builtin
# pylint:disable=invalid-name

# Pylint cannot infer the attributes on world
# pylint:disable=no-member


def string_literal(content):
    """
    Choose a string literal that can wrap our string.

    If your string contains a ``\'`` the result will be wrapped in ``\"``.
    If your string contains a ``\"`` the result will be wrapped in ``\'``.

    Cannot currently handle strings which contain both ``\"`` and ``\'``.
    """

    if '"' in content and "'" in content:
        # there is no way to escape string literal characters in XPath
        raise ValueError("Cannot represent this string in XPath")
    elif '"' in content:  # if it contains " wrap it in '
        content = "'%s'" % content
    else:  # wrap it in "
        content = '"%s"' % content

    return content


class ElementSelector(object):
    """
    A set of elements on a page matching an XPath query.

    :param browser: ``world.browser``
    :param str xpath: XPath query
    :param list elements: list of :class:`selenium.WebElement` objects
    :param bool filter_displayed: whether to only return displayed elements
    :param bool filter_enabled: whether to only return enabled elements

    Delays evaluation to batch the queries together, allowing operations on
    selectors (e.g. union) to be performed first, and then issuing as few
    requests to the browser as possible.

    One of `xpath` or `elements` must be passed. Passing `xpath` creates a
    selector delaying evaluation until it's needed, passing `elements`
    stores the elements immediately.

    Can behave as an iterable of elements or a single element by proxying all
    method calls, asserting that there is only one element selected.

    Can be combined using the addition operator (``+``) to `OR` XPath queries
    together.
    """

    def __init__(self, browser, xpath=None, elements=None,  # pylint:disable=too-many-arguments
                 filter_displayed=False, filter_enabled=False):
        """
        Initialise the selector.

        One of `xpath` or `elements` must be passed. Passing `xpath` creates a
        selector delaying evaluation until it's needed, passing `elements`
        stores the elements immediately.
        """
        self.browser = browser

        if xpath is None and elements is None:
            raise ValueError("Must supply either xpath or elements.")

        if xpath is not None:
            self.xpath = xpath
        else:
            self._elements_cached = elements

        self.filter_displayed = filter_displayed
        self.filter_enabled = filter_enabled

    @property
    def evaluated(self):
        """Whether the selector has already been evaluated."""

        return hasattr(self, '_elements_cached')

    def filter(self, displayed=False, enabled=False):
        """
        Filter elements by visibility and enabled status.

        :param displayed: whether to filter out invisible elements
        :param enabled: whether to filter out disabled elements

        Returns: an :class:`ElementSelector`
        """

        if self.evaluated:
            # Filter elements one by one
            result = self

            if displayed:
                result = ElementSelector(
                    result.browser,
                    elements=[e for e in result if e.is_displayed()]
                )

            if enabled:
                result = ElementSelector(
                    result.browser,
                    elements=[e for e in result if e.is_enabled()]
                )

        else:
            result = copy(self)
            if displayed:
                result.displayed = True
            if enabled:
                result.enabled = True

        return result

    def _select(self):
        """Fetch the elements from the browser."""

        for element in self.browser.find_elements_by_xpath(self.xpath):
            if self.filter_displayed:
                if not element.is_displayed():
                    continue

            if self.filter_enabled:
                if not element.is_enabled():
                    continue

            yield element

    def _elements(self):
        """
        The cached list of elements.
        """
        if not self.evaluated:
            setattr(self, '_elements_cached', list(self._select()))
        return self._elements_cached

    def __add__(self, other):
        """
        Return a union of the two selectors.

        Where possible, avoid evaluating either selector to batch queries.
        """

        if not self.evaluated \
                and isinstance(other, ElementSelector) \
                and not other.evaluated \
                and self.filter_displayed == other.filter_displayed \
                and self.filter_enabled == other.filter_enabled:
            # Both summands are delayed, return a new delayed selector
            return ElementSelector(
                self.browser,
                xpath=self.xpath + '|' + other.xpath,
                filter_displayed=self.filter_displayed,
                filter_enabled=self.filter_enabled,
            )
        else:
            # Have to evaluate everything now
            # other can be either an already evaluated ElementSelector, a list
            # or a single element
            try:
                other = list(other)
            except TypeError:
                other = [other]

            return ElementSelector(self.browser, elements=list(self) + other)

    # The class behaves as a container for the elements, fetching the list from
    # the browser on the first attempt to enumerate itself.

    def __len__(self):
        return len(self._elements())

    def __getitem__(self, key):
        return self._elements()[key]  # pylint:disable=unsubscriptable-object

    def __iter__(self):
        for el in self._elements():  # pylint:disable=not-an-iterable
            yield el

    def __nonzero__(self):
        return bool(self._elements())

    def __getattr__(self, attr):
        """
        Delegate all calls to the only element selected.
        """

        if attr == '_elements_cached':
            # Never going to be on the element
            raise AttributeError()

        assert len(self) == 1, \
            'Must be a single element, have {0}'.format(len(self))
        return getattr(self[0], attr)


def element_id_by_label(browser, label):
    """
    The ID of an element referenced by a `label`s ``for`` attribute. The label
    must be visible.

    :param browser: ``world.browser``
    :param label: label text to return the referenced element for

    Returns: ``for`` attribute value
    """
    label = ElementSelector(browser,
                            str('//label[contains(., %s)]' %
                                string_literal(label)))
    if not label:
        return False
    return label.get_attribute('for')


def field_xpath(field, attribute):
    """
    Field helper functions to locate select, textarea, and the other
    types of input fields (text, checkbox, radio)

    :param field: One of the values 'select', 'textarea', 'option', or
    'button-element' to match a corresponding HTML element (and to
    match a <button> in the case of 'button-element'). Otherwise a
    type to match an <input> element with a type=<field> attribute.

    :param attribute: An attribute to be matched against, or 'value'
    to match against the content within element being matched.
    """
    if field in ['select', 'textarea']:
        xpath = './/{field}[@{attr}=%s]'

    elif field == 'button-role':
        if attribute == 'value':
            xpath = './/*[@role="button"][contains(., %s)]'
        else:
            xpath = './/*[@role="button"][@{attr}=%s]'

    elif field == 'button-element':
        field = 'button'
        if attribute == 'value':
            xpath = './/{field}[contains(., %s)]'
        else:
            xpath = './/{field}[@{attr}=%s]'

    elif field == 'option':
        xpath = './/{field}[@{attr}=%s]'

    else:
        xpath = './/input[@{attr}=%s][@type="{field}"]'

    return xpath.format(field=field, attr=attribute)


def find_button(browser, value):
    """
    Find a button with the given value.

    Searches for the following different kinds of buttons:

        <input type="submit">
        <input type="reset">
        <input type="button">
        <input type="image">
        <button>
        <{a,p,div,span,...} role="button">

    Returns: an :class:`ElementSelector`
    """
    field_types = (
        'submit',
        'reset',
        'button-element',
        'button',
        'image',
        'button-role',
    )

    return reduce(
        operator.add,
        (find_field_with_value(browser, field_type, value)
         for field_type in field_types)
    )


def find_field_with_value(browser, field, value):
    return find_field_by_id(browser, field, value) + \
        find_field_by_name(browser, field, value) + \
        find_field_by_value(browser, field, value)


def find_option(browser, select_name, option_name):
    # First, locate the select
    select_box = find_field(browser, 'select', select_name)
    assert select_box, "Cannot find a '{}' select.".format(select_name)

    # Now locate the option
    option_box = find_field(select_box, 'option', option_name)
    if not option_box:
        # Locate by contents
        option_box = select_box.find_element_by_xpath(str(
            './/option[contains(., "%s")]' % option_name))
    return option_box


def find_field(browser, field_type, value):
    """
    Locate an input field.

    :param browser: ``world.browser``
    :param string field_type: a field type (i.e. `button`)
    :param string value: an id, name or label

    This first looks for `value` as the id of the element, else
    the name of the element, else as a label for the element.

    Returns: an :class:`ElementSelector`
    """
    return find_field_by_id(browser, field_type, value) + \
        find_field_by_name(browser, field_type, value) + \
        find_field_by_label(browser, field_type, value)


def find_any_field(browser, field_types, field_name):
    """
    Find a field of any of the specified types.

    :param browser: ``world.browser``
    :param list field_types: a list of field type (i.e. `button`)
    :param string value: an id, name or label

    Returns: an :class:`ElementSelector`

    See also: :func:`find_field`.
    """

    return reduce(
        operator.add,
        (find_field(browser, field_type, field_name)
         for field_type in field_types)
    )


def find_field_by_id(browser, field_type, id):
    """
    Locate the control input with the given ``id``.

    :param browser: ``world.browser``
    :param string field_type: a field type (i.e. `button`)
    :param string id: ``id`` attribute

    Returns: an :class:`ElementSelector`
    """
    return ElementSelector(
        browser,
        xpath=field_xpath(field_type, 'id') % string_literal(id),
        filter_displayed=True,
    )


def find_field_by_name(browser, field_type, name):
    """
    Locate the control input with the given ``name``.

    :param browser: ``world.browser``
    :param string field_type: a field type (i.e. `button`)
    :param string name: ``name`` attribute

    Returns: an :class:`ElementSelector`
    """
    return ElementSelector(
        browser,
        field_xpath(field_type, 'name') %
        string_literal(name),
        filter_displayed=True,
    )


def find_field_by_value(browser, field_type, name):
    """
    Locate the control input with the given ``value``. Useful for buttons.

    :param browser: ``world.browser``
    :param string field_type: a field type (i.e. `button`)
    :param string name: ``value`` attribute

    Returns: an :class:`ElementSelector`
    """
    xpath = field_xpath(field_type, 'value') % string_literal(name)
    elems = ElementSelector(
        browser,
        xpath=str(xpath),
        filter_displayed=True,
        filter_enabled=True,
    )

    # sort by shortest first (most closely matching)
    if field_type == 'button-element' or field_type == 'button-role':
        elems = sorted(elems, key=lambda elem: len(elem.text))
    else:
        elems = sorted(elems,
                       key=lambda elem: len(elem.get_attribute('value')))

    if elems:
        elems = [elems[0]]  # pylint:disable=redefined-variable-type

    return ElementSelector(browser, elements=elems)


def find_field_by_label(browser, field_type, label):
    """
    Locate the control input that has a label pointing to it.

    :param browser: ``world.browser``
    :param string field_type: a field type (i.e. `button`)
    :param string label: label text

    This will first locate the label element that has a label of the given
    name. It then pulls the id out of the 'for' attribute, and uses it to
    locate the element by its id.

    Returns: an :class:`ElementSelector`
    """

    return ElementSelector(
        browser,
        xpath=field_xpath(field_type, 'id') %
        '//label[contains(., {0})]/@for'.format(
            string_literal(label)),
        filter_displayed=True,
    )


def option_in_select(browser, select_name, option):
    """
    Returns the Element specified by @option or None

    Looks at the real <select> not the select2 widget, since that doesn't
    create the DOM until we click on it.
    """

    select = find_field(browser, 'select', select_name)
    assert select, "Cannot find a '{}' select.".format(select_name)

    try:
        return select.find_element_by_xpath(str(
            './/option[normalize-space(text())=%s]' % string_literal(option)))
    except NoSuchElementException:
        return None


TIMEOUT = 15
CHECK_EVERY = 0.2


def wait_for(func):
    """
    A decorator to invoke a function, retrying on assertion errors for a
    specified time interval.

    Adds a kwarg `timeout` to `func` which is a number of seconds to try
    for (default 15).
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        timeout = kwargs.pop('timeout', TIMEOUT)

        start = None

        while True:
            try:
                return func(*args, **kwargs)
            except AssertionError:
                # The function took some time to test the assertion, however,
                # the result might correspond to the state of the world at any
                # point in time, perhaps earlier than the timeout. Therefore,
                # start counting time from the first assertion fail, not from
                # before the function was called.
                if not start:
                    start = time()
                if time() - start < timeout:
                    sleep(CHECK_EVERY)
                    continue
                else:
                    raise

    return wrapped



"""
Setup script.
"""

from setuptools import setup, find_packages

if __name__ == '__main__':
    with \
            open('requirements.txt') as requirements, \
            open('test_requirements.txt') as test_requirements, \
            open('README.md') as readme:
        setup(
            name='aloe_webdriver',
            use_scm_version=True,
            description='Selenium webdriver extension for Aloe',
            author=", ".join((
                'Alexey Kotlyarov',
                'Danielle Madeley',
                'Nick Pilon',
                'Ben Bangert',
            )),
            author_email=', '.join((
                'a@koterpillar.com',
                'danielle@madeley.id.au',
                'npilon@gmail.com,',
                'ben@groovie.org',
            )),
            url="https://github.com/aloetesting/aloe_webdriver/",
            long_description=readme.read(),
            classifiers=[
                'License :: OSI Approved :: MIT License',
                'Programming Language :: Python',
                'Programming Language :: Python :: 2',
                'Programming Language :: Python :: 3',
            ],

            packages=find_packages(),
            include_package_data=True,

            setup_requires=['setuptools_scm'],

            install_requires=requirements.readlines(),

            test_suite='aloe_webdriver',
            tests_require=test_requirements.readlines(),
        )



"""Test CSS selector steps."""

from aloe.testing import FeatureTest

from aloe_webdriver.tests.base import feature


class TestCSS(FeatureTest):
    """Test CSS steps."""

    @feature()
    def test_css_match(self):
        """
        When I visit test page "basic_page"
        Then there should be an element matching $("textarea[name='bio']")
        """

    @feature(fails=True)
    def test_css_match_failure(self):
        """
        When I visit test page "basic_page"
        Then there should be an element matching $("textarea[name='fail']")
        """

    @feature()
    def test_css_no_match(self):
        """
        When I visit test page "basic_page"
        Then there should not be an element matching $("textarea[name='fail']")
        """

    @feature(fails=True)
    def test_css_no_match_failure(self):
        """
        When I visit test page "basic_page"
        Then there should not be an element matching $("textarea[name='bio']")
        """

    @feature()
    def test_css_count(self):
        """
        When I visit test page "basic_page"
        Then there should be exactly 2 elements matching $("select")
        """

    @feature(fails=True)
    def test_css_count_failure(self):
        """
        When I visit test page "basic_page"
        Then there should be exactly 4 elements matching $("select")
        """

    @feature()
    def test_css_click(self):
        """
        When I visit test page "basic_page"
        And I click $("input[value='Reveal']")
        Then I should see "You pressed an input button"
        """

    @feature()
    def test_forms(self):
        """
        When I visit test page "basic_page"
        Then I fill in $("input[name='user']") with "A test string"
        And I check $("input[value='Bike']")
        And I select $("option[value='green']")
        And I select $("option[value='green']")
        Then $("option[value='green']") should be selected
        And I submit $("form")
        """

    @feature(fails=True)
    def test_select_failure(self):
        """
        When I visit test page "basic_page"
        And I select $("option[value='magenta']")
        """

    @feature(fails=True)
    def test_selected_failure(self):
        """
        When I visit test page "basic_page"
        Then $("option[value='black']") should be selected
        """

    @feature()
    def test_link(self):
        """
        Given I visit test page "link_page"
        And I see "Page o link"
        When I follow the link $("a[href='link_dest.html']")
        Then I should be at "http://SERVER_HOST:7755/link_dest.html"
        """

    @feature()
    def test_delayed_selector(self):
        (
            '''When I visit test page "basic_page"'''
            "\n"
            """And I click $("input[value='Start timer']")\n"""
            """Then there should be an element matching """
            """$("p[id='changed_id']") within 15 seconds"""
        )

    @feature(fails=True)
    def test_delayed_selector_failure(self):
        (
            '''When I visit test page "basic_page"'''
            "\n"
            """And I click $("input[value='Start timer']")\n"""
            """And there should be an element matching """
            """$("p[id='changed_id']") within 1 second"""
        )
